-- by modelleicher
-- December 2016
-- www.schwabenmodding.de
-- facebook.de/modelleicher

-- Purpose:
-- This script is used to outsource all the important functions to attach and detach trees in FS17. 
-- That way all the functions and events for MP functionality can be used by multiple scripts for 
-- different tree attachers within the mod without rewriting all the important attaching stuff.
-- All attachers are saved within the Table self.winchUtilHelper.treeAttacher[attacherIndex] and within this table
-- are some mandatory variables for the functions in the winchUtil to make it work. So the setup of all the attachers is the same.

-- Version 1.0, Initial Version
-- Version 1.1, MP Fix 29-01-2017 - all tree attachers in the same table within "self" in the specializations, winchUtil access via "self" and attacherCountId -> self is mp synchronised so will be attacherCountId

winchUtil = {};

function winchUtil:clientServerDetachTree(self, attacherCountId, noEventSend)
    --print("clientServerDetachTree_called")
	local cTable = self.winchUtilHelper.treeAttacher[attacherCountId]
    if self.isServer then
        setMass(cTable.localAttachedSplit, cTable.splitBackupMass);    
        cTable.splitBackupMass = nil;
        removeJoint(cTable.serverAttachedSplit);
        --print("clientServerDetachTree_called_isServer")
    end;
    cTable.serverAttachedSplit = nil;    
    cTable.localAttachedSplit = nil;
    cTable.localSplitAttacherRef = nil;
    cTable.foundLocalSplit = nil;
    cTable.wUpdated = true;
    cTable.winching = false;
    setVisibility(cTable.rope, false);    

    winchDetachTreeEvent.sendEvent(self, attacherCountId, noEventSend);
end;
--
function winchUtil:createFinderShape(centerNode)
    local planePoint = createTransformGroup("planePoint");
    link(centerNode, planePoint);
    setRotation(planePoint, math.rad(0), math.rad(0), math.rad(180));
    setTranslation(planePoint, 0, 0.5, -0.5);
    
    local dir1 = createTransformGroup("dir1");
    local dir2 = createTransformGroup("dir2");
    link(planePoint, dir1);
    link(planePoint, dir2);
    setTranslation(dir1, 1, 0, 0);
    setTranslation(dir2, 0, 1, 0);
    --print("createFinderShape_created")
    return planePoint, dir1, dir2;
end;

function winchUtil:clientServerFindTree(self, attacherCountId, tX, tY, tZ, rX, rY, rZ, isLocal, noEventSend, lA, lB, lC, scaleY, scaleZ)
	local cTable = self.winchUtilHelper.treeAttacher[attacherCountId];
    --print("ClientServer_findTree called");
    cTable.wUpdated = true;
    -- find the tree for each client/server according to center point pos/rot
    if not isLocal then -- only find tree for others, not for oneself since the current player already found the tree
        --print("clientServer_findTree_notLocal");
        local center = createTransformGroup("local_center");
        link(getRootNode(), center);
        setTranslation(center, tX, tY, tZ);
        setRotation(center, rX, rY, rZ);
        local planePoint, dir1, dir2 = self:createFinderShape(center);
        local pX, pY, pZ = getWorldTranslation(planePoint);
        local dir1X, dir1Y, dir1Z = getWorldTranslation(dir1);
        local dir2X, dir2Y, dir2Z = getWorldTranslation(dir2);
        local nx, ny, nz, yx, yy, yz = dir1X-pX, dir1Y-pY, dir1Z-pZ, dir2X-pX, dir2Y-pY, dir2Z-pZ;    
        local shape, minY, maxY, minZ, maxZ = findSplitShape(pX,pY,pZ, nx, ny, nz, yx, yy, yz, 1, 1);    
        if shape == nil or shape == 0 then
            --print("ERROR: Winch.lua - No shape found at given coordinates!");
        else
            --print("self.w.foundLocalSplit -> found split "..tostring(minY));
            cTable.foundLocalSplit = shape;
            cTable.foundLocalSplitSizes = {minY, maxY, minZ, maxZ};
        
            -- loop placement (and attacher placement relative to that) data
            --local lA, lB, lC = localToWorld(planePoint, 0, (minY+maxY)*0.5, (minZ+maxZ)*0.5)
            --local scaleY = maxY-minY;
            --local scaleZ = maxZ-minZ;
            winchUtil:clientServerAttachTree(self, attacherCountId, tX, tY, tZ, rX, rY, rZ, lA, lB, lC, scaleY, scaleZ);
            
        end;
        delete(center);
    else 
        --print("clientServer_findTree_isLocal");
        winchUtil:clientServerAttachTree(self, attacherCountId, tX, tY, tZ, rX, rY, rZ, lA, lB, lC, scaleY, scaleZ);
        winchAttachTreeEvent.sendEvent(self, attacherCountId, tX, tY, tZ, rX, rY, rZ, noEventSend);
    end;
end;

function winchUtil:clientServerAttachTree(self, attacherCountId, tX, tY, tZ, rX, rY, rZ, lA, lB, lC, scaleY, scaleZ)
    --print("clientServer_attachTree_called")
    local cTable = self.winchUtilHelper.treeAttacher[attacherCountId];
		
    local split = Utils.getNoNil(cTable.foundLocalSplit, cTable.currentLocalSplit);
    --print("attach_tree_split: "..tostring(split));
    cTable.localSplitAttacherRef = createTransformGroup("attacherReference"); -- create attacher reference 
    link(split, cTable.localSplitAttacherRef);        
    local lx, ly, lz = worldToLocal(cTable.localSplitAttacherRef, tX, tY, tZ);    -- position of where the player attaches the rope
    setTranslation(cTable.localSplitAttacherRef, lx, ly, lz);
    cTable.localAttachedSplit = split;
    
    
    -- 
    
    if self.isServer then -- only the server can actually attach the tree
        
        -- lower the mass of the tree while it is attached to stabilize the joint
        cTable.splitBackupMass = getMass(split);
        setMass(split, cTable.splitBackupMass*0.5);
        
        -- lower mass of rope component during attaching to stop tree from jumping
        setMass(self.components[cTable.attacherCompIndex].node, 0.01);
        cTable.massTimer = true;
        
        local constr = JointConstructor:new();
        constr:setActors(self.components[cTable.attacherCompIndex].node, split);
        
        local x, y, z = getWorldTranslation(cTable.attacherNode);
        local a, b, c = getWorldTranslation(cTable.localSplitAttacherRef);
        constr:setJointTransforms(cTable.attacherNode, cTable.localSplitAttacherRef);
        -- get the current distance and set the translation limit to it
        local distance = Utils.vector3Length(x-a, y-b, z-c);
        transRadius = distance+0.1
        local ten = 0.05
        constr:setTranslationLimit(0, true, -ten, ten);
        constr:setTranslationLimit(1, true, -ten, ten);
        constr:setTranslationLimit(2, true, -transRadius, transRadius);

        constr:setTranslationLimitSpring(0, 0, 0, 0, 0, 0);
        
        constr:setEnableCollision(false);
        cTable.serverAttachedSplit = constr:finalize();    
        setJointRotationLimit(cTable.serverAttachedSplit, 0, false, 0, 0); -- set rot limit false since the tree is able to rotate on the winch
        cTable.curLimit = transRadius+0.01;
        

        --print("done attaching");
    end;    
    cTable.foundLocalSplit = nil;
end;




winchAttachTreeEvent = {};
winchAttachTreeEvent_mt = Class(winchAttachTreeEvent, Event);

InitEventClass(winchAttachTreeEvent, "winchAttachTreeEvent");

function winchAttachTreeEvent:emptyNew()
    local self = Event:new(winchAttachTreeEvent_mt);
    return self;
end;

function winchAttachTreeEvent:new(object, attacherCountId, tX, tY, tZ, rX, rY, rZ)
    local self = winchAttachTreeEvent:emptyNew();
    self.object = object;
	self.attacherCountId = attacherCountId;
	--print(" - DEBUG - new - cTable: "..tostring(attacherCountId));
    self.tX = tX;
    self.tY = tY;
    self.tZ = tZ;
    self.rX = rX;
    self.rY = rY;
    self.rZ = rZ;
    return self;
end;

function winchAttachTreeEvent:readStream(streamId, connection)
    local id = streamReadInt32(streamId);
    self.object = networkGetObject(id);
	self.attacherCountId = streamReadInt32(streamId);
    self.tX = streamReadFloat32(streamId);
    self.tY = streamReadFloat32(streamId);
    self.tZ = streamReadFloat32(streamId);
    self.rX = streamReadFloat32(streamId);
    self.rY = streamReadFloat32(streamId);
    self.rZ = streamReadFloat32(streamId);    
	
	if self.object ~= nil then
		--print(" - DEBUG - readStream - attacherCountId: "..tostring(self.attacherCountId));
	else
		--print(" - DEBUG - readStream - self.object == nil");
	end;	
    self:run(connection);
end;

function winchAttachTreeEvent:writeStream(streamId, connection)
    streamWriteInt32(streamId, networkGetObjectId(self.object));
	--print(" - DEBUG - writeStream - attacherCountId: "..tostring(self.attacherCountId));
	streamWriteInt32(streamId, self.attacherCountId);
    streamWriteFloat32(streamId, self.tX);
    streamWriteFloat32(streamId, self.tY);
    streamWriteFloat32(streamId, self.tZ);
    streamWriteFloat32(streamId, self.rX);
    streamWriteFloat32(streamId, self.rY);    
    streamWriteFloat32(streamId, self.rZ);
end;

function winchAttachTreeEvent:run(connection)
    --print("EVENT_winchAttachTreeEvent_RUN");
	--print(" - DEBUG - run - attacherCountId: "..tostring(self.attacherCountId));
	--if self.object ~= nil then
	if not connection:getIsServer() then
        g_server:broadcastEvent(winchAttachTreeEvent:new(self.object, self.attacherCountId, self.tX, self.tY, self.tZ, self.rX, self.rY, self.rZ), nil, connection, self.object);
    end;
	
    winchUtil:clientServerFindTree(self.object, self.attacherCountId, self.tX, self.tY, self.tZ, self.rX, self.rY, self.rZ, false, true);

end;

function winchAttachTreeEvent.sendEvent(object, attacherCountId, tX, tY, tZ, rX, rY, rZ, noEventSend)
    --print("EVENT_winchAttachTreeEvent_sendEvent");
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            --print("EVENT_winchAttachTreeEvent_sendEvent_g_server");
            g_server:broadcastEvent(winchAttachTreeEvent:new(object, attacherCountId, tX, tY, tZ, rX, rY, rZ ), nil, nil, object);
			--print(" - DEBUG - sendEvent - gServer ~= nil - attacherCountId: "..tostring(attacherCountId));
        else
            --print("EVENT_winchAttachTreeEvent_sendEvent_g_client");
            g_client:getServerConnection():sendEvent(winchAttachTreeEvent:new(object, attacherCountId, tX, tY, tZ, rX, rY, rZ));
			--print(" - DEBUG - sendEvent - gServer == nil - attacherCountId: "..tostring(attacherCountId));
        end;
    end;
end;

                
winchDetachTreeEvent = {};
winchDetachTreeEvent_mt = Class(winchDetachTreeEvent, Event);

InitEventClass(winchDetachTreeEvent, "winchDetachTreeEvent");

function winchDetachTreeEvent:emptyNew()
    local self = Event:new(winchDetachTreeEvent_mt);
    return self;
end;

function winchDetachTreeEvent:new(object, attacherCountId)
    local self = winchDetachTreeEvent:emptyNew();
    self.object = object;
    self.attacherCountId = attacherCountId;
    return self;
end;

function winchDetachTreeEvent:readStream(streamId, connection)
    local id = streamReadInt32(streamId);
    self.object = networkGetObject(id);
	self.attacherCountId = streamReadInt32(streamId);
    self:run(connection);
end;

function winchDetachTreeEvent:writeStream(streamId, connection)
    streamWriteInt32(streamId, networkGetObjectId(self.object));
    streamWriteInt32(streamId, self.attacherCountId);

end;

function winchDetachTreeEvent:run(connection)
    --print("EVENT_winchDetachTreeEvent_run");
    if not connection:getIsServer() then
        g_server:broadcastEvent(winchDetachTreeEvent:new(self.object, self.attacherCountId), nil, connection, self.object);
    end;	
	
    winchUtil:clientServerDetachTree(self.object, self.attacherCountId, true);

end;
    
function winchDetachTreeEvent.sendEvent(object, attacherCountId, noEventSend)
    --print("EVENT_winchDetachTreeEvent_sendEvent");
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            --print("EVENT_winchDetachTreeEvent_sendEvent_g_server");
            --print("EVENT_winchDetachTreeEvent_sendEvent_g_client");
			g_server:broadcastEvent(winchDetachTreeEvent:new(object, attacherCountId), nil, nil, object);
        else
            g_client:getServerConnection():sendEvent(winchDetachTreeEvent:new(object, attacherCountId));
        end;
    end;
end;    
